#!/bin/bash

# Create certificate for Server Authentication and Client Authentication

# Example for use with hostname and FQDN:
# ./create_certificate.sh --dns=apmaccs,apmaccs.sos --days=7300

set -e

CA_HOME=$(cd "$(dirname "$0")" >/dev/null && pwd)
CA_CERTS="${CA_HOME}"/certs
CA_CSR="${CA_HOME}"/csr
CA_PRIVATE="${CA_HOME}"/private
CA_CONFIG="${CA_HOME}"/openssl-cert.config

# default certificate lifetime
days=5475

# Check command line options
for option in "$@"
do
    case "${option}" in
            --dns=*)   dns=$(echo "${option}" | sed 's/--dns=//' | sed 's/^"//' | sed 's/"$//' | sed -r 's/[,]+/ /g')
                       ;;
            --days=*)  days=$(echo "${option}" | sed 's/--days=//' | sed 's/^"//' | sed 's/"$//' | sed 's/^\(.*\)\/$/\1/')
                       ;;
            *)         >&2 echo "unknown option: ${option}, use --dns="
                       exit 1
                       ;;
    esac
done

if [ -z "${dns}" ]; then
    echo "DNS name missisng for server, use: --dns=<hostname> or --dns=<hostname>,<fqdn>"
    exit 1
fi

dnsName=
dnsNames=
for s in ${dns[@]}; do
    if [ -z "${dnsName}" ]; then
        dnsName=${s}
    else
        dnsNames="${dnsNames}",
    fi
    
    dnsNames=${dnsNames}DNS:${s}
done

# Create required sub-directories
mkdir -p "${CA_CERTS}" "${CA_CSR}" "${CA_PRIVATE}"

# Update openssl-cert.config for hostnane
sed -i "s/^[# ]*commonName[ ]*=.*/            commonName = ${dnsName}/g" "${CA_CONFIG}"

# Step 1 - Generate Private Key and Certificate Signing Request
openssl req -new -config "${CA_CONFIG}" -extensions 'standard exts' -nodes \
    -days "${days}" -newkey rsa:4096 -keyout "${CA_PRIVATE}"/"${dnsName}".key -out "${CA_CSR}"/"${dnsName}".csr

# Step 2 - Generate and sign the Server Certificate
openssl x509 -req \
    -in "${CA_CSR}"/"${dnsName}".csr \
    -CA "${CA_CERTS}"/root-ca.crt \
    -CAkey "${CA_PRIVATE}"/root-ca.key \
    -CAcreateserial \
    -out "${CA_CERTS}"/"${dnsName}".crt -days "${days}" \
    -extfile <(printf 'subjectAltName=%s\nnsCertType = client, server\nkeyUsage =  critical, nonRepudiation, digitalSignature, keyEncipherment\nextendedKeyUsage = serverAuth, clientAuth\n' "${dnsNames}")

